package openapi

import (
	"fmt"
	. "github.com/smartystreets/goconvey/convey"
	"io/ioutil"
	"os"
	"testing"
)

func TestServiceSchemaConfigurationV1(t *testing.T) {
	Convey("Given a SchemaPropertyName, DefaultValue and some ExternalConfiguration ", t, func() {
		schemaPropertyName := "schemaPropertyName"
		defaultValue := "defaultValue"
		externalConfiguration := ServiceSchemaPropertyExternalConfigurationV1{
			KeyName:     "someKeyName",
			ContentType: "raw",
			File:        "/path/to/credentials",
		}
		Convey("When a new ServiceSchemaPropertyConfigurationV1 is created", func() {
			serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
				SchemaPropertyName:    schemaPropertyName,
				DefaultValue:          defaultValue,
				ExternalConfiguration: externalConfiguration,
			}
			Convey("And the serviceSchemaConfigurationV1 created should implement ServiceSchemaPropertyConfiguration interface", func() {
				var _ ServiceSchemaPropertyConfiguration = serviceSchemaConfigurationV1
			})
		})
	})
}

func TestServiceSchemaConfigurationV1ExecuteCommand(t *testing.T) {
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that exists successfully) configured", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"date"},
		}
		Convey("When ExecuteCommand method is called", func() {
			err := serviceSchemaConfigurationV1.ExecuteCommand()
			Convey("And the err returned should be nil", func() {
				So(err, ShouldBeNil)
			})
		})
	})
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that exists with error) configured", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"cat", "nonexistingfile"},
		}
		Convey("When ExecuteCommand method is called", func() {
			err := serviceSchemaConfigurationV1.ExecuteCommand()
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message returned should be the expected", func() {
				So(err.Error(), ShouldEqual, "provider schema property 'some_property_name' command failed: command '[cat nonexistingfile]' failed: cat: nonexistingfile: No such file or directory\n(exit status 1)")
			})
		})
	})
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that timeouts) configured", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"sleep", "2"},
			CommandTimeout:     1,
		}
		Convey("When ExecuteCommand method is called", func() {
			err := serviceSchemaConfigurationV1.ExecuteCommand()
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message returned should be the expected", func() {
				So(err.Error(), ShouldEqual, "provider schema property 'some_property_name' command failed: command '[sleep 2]' did not finish executing within the expected time 1s (signal: killed)")
			})
		})
	})
}

func TestServiceSchemaConfigurationV1Exec(t *testing.T) {
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that exists successfully) configured and a channel", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"date"},
		}
		doneChan := make(chan error)
		Convey("When exec method is called", func() {
			go serviceSchemaConfigurationV1.exec(doneChan)
			Convey("And the err returned should be nil", func() {
				err := <-doneChan
				So(err, ShouldBeNil)
			})
		})
	})
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that exists with error) configured and a channel", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"cat", "nonexistingfile"},
		}
		doneChan := make(chan error)
		Convey("When exec method is called", func() {
			go serviceSchemaConfigurationV1.exec(doneChan)
			err := <-doneChan
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message returned should be the expected", func() {
				So(err.Error(), ShouldEqual, "command '[cat nonexistingfile]' failed: cat: nonexistingfile: No such file or directory\n(exit status 1)")
			})
		})
	})
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with a command (that timeouts) and a channel", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "some_property_name",
			Command:            []string{"sleep", "2"},
			CommandTimeout:     1,
		}
		doneChan := make(chan error)
		Convey("When exec method is called", func() {
			go serviceSchemaConfigurationV1.exec(doneChan)
			err := <-doneChan
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message returned should be the expected", func() {
				So(err.Error(), ShouldEqual, "command '[sleep 2]' did not finish executing within the expected time 1s (signal: killed)")
			})
		})
	})
}

func TestServiceSchemaConfigurationV1GetDefaultValue(t *testing.T) {
	Convey("Given a ServiceSchemaPropertyConfigurationV1 with just a default value and no external config", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
		}
		Convey("When GetDefaultValue method is called", func() {
			value, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the value returned should matched the default one", func() {
				So(value, ShouldEqual, serviceSchemaConfigurationV1.DefaultValue)
			})
		})
	})

	Convey("Given an external config file and a ServiceSchemaPropertyConfigurationV1 with a default value and external 'raw' config", t, func() {
		expectedValue := "some content"
		tmpFile, err := ioutil.TempFile("", "")
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(expectedValue))
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
			ExternalConfiguration: ServiceSchemaPropertyExternalConfigurationV1{
				ContentType: "raw",
				File:        tmpFile.Name(),
			},
		}
		Convey("When GetDefaultValue method is called", func() {
			value, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the value returned should matched the default one", func() {
				So(value, ShouldEqual, expectedValue)
			})
		})
	})

	Convey("Given an external config file and a ServiceSchemaPropertyConfigurationV1 with a default value and external 'json' config", t, func() {
		expectedValue := "someName"
		tmpFile, err := ioutil.TempFile("", "")
		defer os.Remove(tmpFile.Name())
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(fmt.Sprintf(`{"firstName":"%s"}`, expectedValue)))
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
			ExternalConfiguration: ServiceSchemaPropertyExternalConfigurationV1{
				KeyName:     "$.firstName",
				ContentType: "json",
				File:        tmpFile.Name(),
			},
		}
		Convey("When GetDefaultValue method is called", func() {
			value, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the value returned should matched the default one", func() {
				So(value, ShouldEqual, expectedValue)
			})
		})
	})

	Convey("Given an external config file and a ServiceSchemaPropertyConfigurationV1 with an external configuration's non supported content type", t, func() {
		expectedValue := "some content"
		tmpFile, err := ioutil.TempFile("", "")
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(expectedValue))
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
			ExternalConfiguration: ServiceSchemaPropertyExternalConfigurationV1{
				ContentType: "nonSupported",
				File:        tmpFile.Name(),
			},
		}
		Convey("When GetDefaultValue method is called", func() {
			_, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message should be", func() {
				So(err.Error(), ShouldContainSubstring, "'schemaPropertyName': 'nonSupported' content type not supported")
			})
		})
	})

	Convey("Given an external config file and a ServiceSchemaPropertyConfigurationV1 with an external configuration pointing a non existing file", t, func() {
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
			ExternalConfiguration: ServiceSchemaPropertyExternalConfigurationV1{
				ContentType: "nonSupported",
				File:        "som_non_existing_file",
			},
		}
		Convey("When GetDefaultValue method is called", func() {
			_, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message should be", func() {
				So(err.Error(), ShouldContainSubstring, "failed to read external configuration file 'som_non_existing_file' for schema property 'schemaPropertyName': open som_non_existing_file: no such file or directory")
			})
		})
	})

	Convey("Given an external config file and a ServiceSchemaPropertyConfigurationV1 with an external configuration pointing a json file but the keyName is wrong", t, func() {
		expectedValue := "someName"
		tmpFile, err := ioutil.TempFile("", "")
		defer os.Remove(tmpFile.Name())
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(fmt.Sprintf(`{"firstName":"%s"}`, expectedValue)))
		serviceSchemaConfigurationV1 := ServiceSchemaPropertyConfigurationV1{
			SchemaPropertyName: "schemaPropertyName",
			DefaultValue:       "defaultValue",
			ExternalConfiguration: ServiceSchemaPropertyExternalConfigurationV1{
				KeyName:     "$.wrong",
				ContentType: "json",
				File:        tmpFile.Name(),
			},
		}
		Convey("When GetDefaultValue method is called", func() {
			_, err := serviceSchemaConfigurationV1.GetDefaultValue()
			Convey("And the err returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message should be", func() {
				So(err.Error(), ShouldContainSubstring, "key error: wrong not found in object")
			})
		})
	})
}

func TestServiceExternalConfigurationV1GetFileParser(t *testing.T) {
	Convey("Given a ServiceSchemaPropertyExternalConfigurationV1 configured with 'raw' content", t, func() {
		expectedValue := "some content"
		tmpFile, err := ioutil.TempFile("", "")
		defer os.Remove(tmpFile.Name())
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(expectedValue))
		serviceExternalConfigurationV1 := ServiceSchemaPropertyExternalConfigurationV1{
			ContentType: "raw",
			File:        tmpFile.Name(),
		}
		Convey("When getFileParser method is called with some content", func() {
			parser, err := serviceExternalConfigurationV1.getFileParser()
			Convey("Then the error returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the parser returned should be of type parserRaw", func() {
				So(parser, ShouldHaveSameTypeAs, parserRaw{})
			})
			Convey("And the parser get value should return the right value", func() {
				value, err := parser.getValue()
				So(err, ShouldBeNil)
				So(value, ShouldEqual, expectedValue)
			})
		})
	})

	Convey("Given a ServiceSchemaPropertyExternalConfigurationV1 configured with 'json' content", t, func() {
		expectedValue := "someName"
		tmpFile, err := ioutil.TempFile("", "")
		tmpFile.Write([]byte(fmt.Sprintf(`{"firstName":"%s"}`, expectedValue)))
		defer os.Remove(tmpFile.Name())
		So(err, ShouldBeNil)
		serviceExternalConfigurationV1 := ServiceSchemaPropertyExternalConfigurationV1{
			ContentType: "json",
			File:        tmpFile.Name(),
			KeyName:     "$.firstName",
		}
		Convey("When getFileParser method is called with some content", func() {
			parser, err := serviceExternalConfigurationV1.getFileParser()
			Convey("Then the error returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the parser returned should be of type parserRaw", func() {
				So(parser, ShouldHaveSameTypeAs, parserJSON{})
			})
			Convey("And the parser get value should return the right value", func() {
				value, err := parser.getValue()
				So(err, ShouldBeNil)
				So(value, ShouldEqual, expectedValue)
			})
		})
	})

	Convey("Given a ServiceSchemaPropertyExternalConfigurationV1 configured with a non supported content", t, func() {
		expectedValue := "some content"
		tmpFile, err := ioutil.TempFile("", "")
		defer os.Remove(tmpFile.Name())
		So(err, ShouldBeNil)
		tmpFile.Write([]byte(expectedValue))
		serviceExternalConfigurationV1 := ServiceSchemaPropertyExternalConfigurationV1{
			ContentType: "nonSupported",
			File:        tmpFile.Name(),
		}
		Convey("When getFileParser method is called with some content", func() {
			_, err := serviceExternalConfigurationV1.getFileParser()
			Convey("Then the error returned should be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the err message should be", func() {
				So(err.Error(), ShouldContainSubstring, "'nonSupported' content type not supported")
			})
		})
	})
}

func TestSchemaFileParserRaw(t *testing.T) {
	Convey("Given a content", t, func() {
		content := "someContent"
		Convey("When a new parserRaw is created", func() {
			parserRaw := parserRaw{
				content: content,
			}
			Convey("And the parserRaw created should implement schemaFileParser interface", func() {
				var _ schemaFileParser = parserRaw
			})
		})
	})
}

func TestSchemaFileParserRawGetValue(t *testing.T) {
	Convey("Given a content", t, func() {
		expectedContent := "some content"
		parserRaw := parserRaw{
			content: expectedContent,
		}
		Convey("When getValue method is called", func() {
			value, err := parserRaw.getValue()
			Convey("Then the error returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the value returned should match the content", func() {
				So(value, ShouldEqual, expectedContent)
			})
		})
	})
}

func TestSchemaFileParserJSON(t *testing.T) {
	Convey("Given a content", t, func() {
		jsonContent := "{}"
		keyName := "$.firstName"
		Convey("When a new parserRaw is created", func() {
			parserJSON := parserJSON{
				jsonContent: jsonContent,
				keyName:     keyName,
			}
			Convey("And the parserJSON created should implement schemaFileParser interface", func() {
				var _ schemaFileParser = parserJSON
			})
		})
	})
}

func TestSchemaFileParserJsonGetValue(t *testing.T) {
	Convey("Given a jsonContent and an existing key name", t, func() {
		expectedValue := "someName"
		jsonContent := fmt.Sprintf(`{"firstName":"%s"}`, expectedValue)
		keyName := "$.firstName"
		parserJSON := parserJSON{
			jsonContent: jsonContent,
			keyName:     keyName,
		}
		Convey("When getValue method is called", func() {
			value, err := parserJSON.getValue()
			Convey("Then the error returned should be nil", func() {
				So(err, ShouldBeNil)
			})
			Convey("And the value returned should match the content", func() {
				So(value, ShouldEqual, expectedValue)
			})
		})
	})

	Convey("Given a content and NON existing key name", t, func() {
		expectedValue := "someName"
		jsonContent := fmt.Sprintf(`{"firstName":"%s"}`, expectedValue)
		keyName := "$.someNonExistingKey"
		parserJSON := parserJSON{
			jsonContent: jsonContent,
			keyName:     keyName,
		}
		Convey("When getValue method is called", func() {
			_, err := parserJSON.getValue()
			Convey("Then the error returned should NOT be nil", func() {
				So(err, ShouldNotBeNil)
			})
			Convey("And the error returned should be", func() {
				So(err.Error(), ShouldEqual, "key error: someNonExistingKey not found in object")
			})
		})
	})

	Convey("Given a wrong json", t, func() {
		jsonContent := `{`
		keyName := "$.someNonExistingKey"
		parserJSON := parserJSON{
			jsonContent: jsonContent,
			keyName:     keyName,
		}
		Convey("When getValue method is called", func() {
			_, err := parserJSON.getValue()
			Convey("Then the error returned should tell that the input is not json", func() {
				So(err.Error(), ShouldEqual, "unexpected end of JSON input")
			})
		})
	})
}
